//
// Programmer:    Craig Stuart Sapp <craig@ccrma.stanford.edu>
// Creation Date: Wed Jan 21 22:46:30 GMT-0800 1998
// Last Modified: Thu Jan 22 22:53:53 GMT-0800 1998
// Filename:      ...sig/code/control/MidiInPort/visual/MidiInPort_visual.cpp
// Syntax:        C++ 
// $Smake:        cc -g -c %b.cpp -I../../../../include && rm -f %b.o
//

#ifdef VISUAL


#include "MidiInPort_visual.h"

#include <windows.h>
#include <mmsystem.h>

#include <iostream.h>


// initialized static variables
int       MidiInPort_visual::numDevices      = 0;
int       MidiInPort_visual::objectCount     = 0;
int*      MidiInPort_visual::openQ           = NULL;
int*      MidiInPort_visual::inrunningQ      = NULL;
int*      MidiInPort_visual::portObjectCount = NULL;
HMIDIIN*  MidiInPort_visual::device          = NULL;
HANDLE*   MidiInPort_visual::hMutex          = NULL;
CircularBuffer<MidiMessage>* MidiInPort_visual::midiBuffer = NULL;



//////////////////////////////
// 
// MidiInPort_visual::MidiInPort_visual
//	default values: autoOpen = 1
//

MidiInPort_visual::MidiInPort_visual(void) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(0);
}


MidiInPort_visual::MidiInPort_visual(int aPort, int autoOpen) {
   if (objectCount == 0) {
      initialize();
   }
   objectCount++;

   port = -1;
   setPort(aPort);
   if (autoOpen) {
      open();
   }
}



//////////////////////////////
//
// MidiInPort_visual::~MidiInPort_visual
//

MidiInPort_visual::~MidiInPort_visual() {
   objectCount--;
   if (objectCount == 0) {
      deinitialize();
   } else if (objectCount < 0) {
      cerr << "Error: bad MidiInPort_visual object count!: " 
           << objectCount << endl;
      exit(1);
   }
}



//////////////////////////////
//
// MidiInPort_visual::close
//

void MidiInPort_visual::close(void) {
   if (getPortStatus() == 1 && device[getPort()] != NULL) {
      midiInReset(device[getPort()]);
      midiInClose(device[getPort()]);
      openQ[getPort()] = 0;
      inrunningQ[getPort()] = 0;
   }
}



//////////////////////////////
//
// MidiInPort_visual::closeAll
//

void MidiInPort_visual::closeAll(void) {
   for (int i=0; i<getNumPorts(); i++) {
      if (openQ[i] == 1 && device[i] != NULL) {
         midiInReset(device[i]);
         midiInClose(device[i]);
         openQ[i] = 0;
         inrunningQ[i] = 0;
      }
   }
}



//////////////////////////////
//
// MidiInPort_visual::extract -- returns the next MIDI message
//	received since that last extracted message.
//

MidiMessage MidiInPort_visual::extract(void) {
   MidiMessage output;
   
   waitForMutex();
   output = midiBuffer[getPort()].extract();
   releaseMutex();

   return output;
}



//////////////////////////////
//
// MidiInPort_visual::getCount -- returns the number of unexamined
//	MIDI messages waiting in the input buffer.
//

int MidiInPort_visual::getCount(void) {
   int output;
  
   waitForMutex();
   output = midiBuffer[getPort()].getCount();
   releaseMutex();

   return output;
}



//////////////////////////////
//
// MidiInPort_visual::getName -- returns the name of the port.
//	returns "" if no name. Name is valid until getName is called again.
//

const char* MidiInPort_visual::getName(void) {
   static MIDIINCAPS inputCapabilities;
   if (openQ[getPort()]) {  // port already open
      midiInGetDevCaps(getPort(), &inputCapabilities, sizeof(MIDIINCAPS));
   } else {  // port is currently closed
      if(open()) {;
         midiInGetDevCaps(getPort(), &inputCapabilities, sizeof(MIDIINCAPS));
         close();
      } else {
         return "";
      }
   }
   return inputCapabilities.szPname;
}



//////////////////////////////
//
// MidiInPort_visual::getNumPorts -- returns the number of available
// 	ports for MIDI input
//

int MidiInPort_visual::getNumPorts(void) {
   return numDevices;
}



//////////////////////////////
//
// MidiInPort_visual::getPort -- returns the port to which this
//	object belongs (as set with the setPort function).
//

int MidiInPort_visual::getPort(void) {
   return port;
}



//////////////////////////////
//
// MidiInPort_visual::getPortStatus -- 0 if closed, 1 if open
//

int MidiInPort_visual::getPortStatus(void) {
   if (openQ[getPort()] == 1) {
      return 1;
   } else {
      return 0;
   }
}



//////////////////////////////
//
// MidiInPort_visual::getTrace -- returns true if trace is on or false
//	if trace is off.  if trace is on, then prints to standard
// 	output the Midi message received.
//

int MidiInPort_visual::getTrace(void) {
   return trace;
}



//////////////////////////////
//
// MidiInPort_visual::insert
//

void MidiInPort_visual::insert(const MidiMessage& aMessage) {
   waitForMutex();
   midiBuffer[getPort()].insert(aMessage);
   releaseMutex();
}



//////////////////////////////
//
// MidiInPort_visual::message
//

MidiMessage& MidiInPort_visual::message(int index) {
   return midiBuffer[getPort()][index];
}



//////////////////////////////
//
// MidiInPort_visual::open -- returns true if MIDI input port was
//	opened.
//

int MidiInPort_visual::open(void) {
   if (getPortStatus() == 0) {
      int flag;
      flag = midiInOpen(&device[getPort()], getPort(), 
         (DWORD)&midiInputCallback, (DWORD)this, CALLBACK_FUNCTION);
      if (flag == MMSYSERR_NOERROR) {
         openQ[getPort()] = 1;
         unpause();
         return 1;
      } else { // failed to open
         openQ[getPort()] = 0;
         device[getPort()] = NULL;
         return 0;
      }
   } else { // already open
      return 1;
   }
}



//////////////////////////////
//
// MidiInPort_visual::pause -- stop the Midi input port from
//	inserting MIDI messages into the buffer, but keep the
//	port open.  Use unpause() to reverse the effect of pause().
//

void MidiInPort_visual::pause(void) {
   if (openQ[getPort()]) {
      if (inrunningQ[getPort()] == 1) {
         midiInStop(device[getPort()]);
         inrunningQ[getPort()] = 0;
      }
   }
}



//////////////////////////////
//
// MidiInPort_visual::setBufferSize -- sets the allocation
//	size of the MIDI input buffer.
//

void MidiInPort_visual::setBufferSize(int aSize) {
   midiBuffer[getPort()].setSize(aSize);
}



//////////////////////////////
//
// MidiInPort_visual::setPort
//

void MidiInPort_visual::setPort(int aPort) {
   if (aPort < 0 || aPort >= getNumPorts()) {
      cerr << "Error: maximum port number is: " << getNumPorts()-1
           << ", but you tried to access port: " << aPort << endl;
      exit(1);
   }

   if (port != -1) {
      portObjectCount[port]--;
   }
   port = aPort;
   portObjectCount[port]++;
}



//////////////////////////////
//
// MidiInPort_visual::setTrace -- if false, then don't print MIDI messages
// 	to the screen.
//

int MidiInPort_visual::setTrace(int aState) {
   int oldtrace = trace;
   if (aState == 0) {
      trace = 0;
   } else {
      trace = 1;
   }
   return oldtrace;
}



//////////////////////////////
//
// MidiInPort_visual::toggleTrace -- switches the state of trace
//	Returns the previous value of the trace variable.
//

void MidiInPort_visual::toggleTrace(void) {
   trace = !trace;
}
   


//////////////////////////////
//
// MidiInPort_visual::unpause -- enables the Midi input port 
//	to inserting MIDI messages into the buffer after the 
//	port is already open.
//

void MidiInPort_visual::unpause(void) {
   if (openQ[getPort()]) {
      if (inrunningQ[getPort()] == 0) {
         midiInStart(device[getPort()]);
         inrunningQ[getPort()] = 1;
      }
   }
}



///////////////////////////////////////////////////////////////////////////
//
// Private functions
//



//////////////////////////////
//
// MidiInPort_visual::deinitialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiInPort_visual::deinitialize(void) {
   closeAll();
   if (device != NULL) delete [] device;
   device = NULL;
   if (openQ != NULL) delete [] openQ;
   openQ = NULL;
   if (inrunningQ != NULL) delete [] inrunningQ;
   inrunningQ = NULL;
   if (hMutex != NULL) delete [] hMutex;
   hMutex = NULL;
   if (midiBuffer != NULL) delete [] midiBuffer;
   midiBuffer = NULL;
   if (portObjectCount != NULL) delete [] portObjectCount;
   portObjectCount = NULL;
}



//////////////////////////////
//
// MidiInPort_visual::initialize -- sets up storage if necessary
//	This function should be called if the current object is
//	the first object to be created.
//

void MidiInPort_visual::initialize(void) {
   // get the number of ports
   numDevices = midiInGetNumDevs();
   if  (getNumPorts() <= 0) {
      cerr << "Error: no MIDI input devices" << endl;
      exit(1);
   }

   // allocate space for Windoze MIDI input structures
   if (device != NULL) {
      cerr << "Error: device array should be NULL when calling "
           << "initialize() in MidiInPort_visual." << endl;
      exit(1);
   }
   device = new HMIDIIN[numDevices];

   // allocate space for openQ, the port open/close status
   if (openQ != NULL) delete [] openQ;
   openQ = new int[numDevices];

   // allocate space to keep track of an active/inactive input port
   if (inrunningQ != NULL) delete [] inrunningQ;
   inrunningQ = new int[numDevices];

   // allocate space for object count on each port:
   if (portObjectCount != NULL) delete [] portObjectCount;
   portObjectCount = new int[numDevices];

   // allocate space for mutual exclusive
   if (hMutex != NULL) delete [] hMutex;
   hMutex = new HANDLE[numDevices];

   // allocate space for the Midi input buffers
   if (midiBuffer != NULL) delete [] midiBuffer;
   midiBuffer = new CircularBuffer<MidiMessage>[numDevices];

   // initialize the static arrays
   for (int i=0; i<getNumPorts(); i++) {
      device[i] = NULL;
      openQ[i] = 0;
      inrunningQ[i] = 1;
      portObjectCount[i] = 0;
      hMutex[i] = CreateMutex(NULL, FALSE, "M");
      midiBuffer[i].setSize(DEFAULT_INPUT_BUFFER_SIZE);
   }
}



//////////////////////////////
//
// MidiInPort_visual::releaseMutex
//

void MidiInPort_visual::releaseMutex(void) {
   int flag = ReleaseMutex(hMutex[getPort()]);
/* 
   if (flag != 0) {
      cerr << "Error relasing mutex in MIDI input; flag was: " << flag << endl;
   }
*/
}



//////////////////////////////
//
// MidiInPort_visual::setPortStatus
//

void MidiInPort_visual::setPortStatus(int aStatus) {
   if (aStatus) {
      openQ[getPort()] = 1;
   } else {
      openQ[getPort()] = 0;
   }
}



//////////////////////////////
//
// MidiInPort_visual::waitForMutex 
//

void MidiInPort_visual::waitForMutex(void) {
   DWORD mutexResult = WaitForSingleObject(hMutex[getPort()], 5000L);
   if (mutexResult != WAIT_OBJECT_0) {
      cerr << "Error waiting for mutex in MIDI input" << endl;
   }
}



///////////////////////////////////////////////////////////////////////////


//////////////////////////////
//
// midiInputCallback -- the function the MIDI input driver calls when
// 	it has a message from the Midi in cable ready
//

void CALLBACK midiInputCallback(HMIDIIN hMidiIn, UINT inputStatus,
      DWORD instancePtr, DWORD midiMessage, DWORD timestamp) {
   MidiMessage newMessage;
   
   switch (inputStatus) {
      case MIM_MOREDATA:
         // there is more data waiting at the device

      case MIM_DATA:
         // ignore the Yamaha Active Sensing command:
         if ((midiMessage & 0xff) == 0xfe || (midiMessage & 0xff) == 0xf8) {
            break;
         }
         newMessage.time = timestamp;
         newMessage.data = midiMessage;
         ((MidiInPort_visual*)instancePtr)->insert(newMessage);
         if (((MidiInPort_visual*)instancePtr)->getTrace()) {
            cout << "[" << hex << (int)newMessage.command() << dec
                 << ":" << (int)newMessage.p1() << ","
                 << (int)newMessage.p2() << "]";
            cout.flush();
         }
         break;
      default: ;
         // must be some sort of error
   }
}



#endif  // VISUAL


